/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pim/mail/MailSyncSource.h"
#include "spds/SyncItemStatus.h"
#include "spds/MailMessage.h"
#include "spds/constants.h"
#include "base/util/utils.h"
#include "base/util/StringBuffer.h"
#include "pim/utils.h"
#include "spds/FolderData.h"
#include "notify/timed_msgbox.h"

#define ASK_REFRESH_MESSAGE TEXT("All the local mails will be deleted before adding the server email. Are you sure to continue?")
#define ASK_REFRESH_TIMEOUT 10

static int call;
static int cnew;
static int cupdated;
static int cdeleted;

static SyncItem *buildSyncItem(MailClientData* m, ArrayList& extras)
{
    MailMessage& msg = m->getEmailData()->getEmailItem();

    SyncItem *syncItem = new SyncItem(m->getEntryId());

    //LOG.debug("MailSyncSource: From: %s", m->getEmailData()->getEmailItem().getFrom());   // TODO: remove this log
    char* toSend = m->getEmailData()->format();
    //LOG.debug(toSend);   // TODO: remove this log
    syncItem->setData(toSend, strlen(toSend) * sizeof(char));              
    syncItem->setSourceParent(m->getParent());
    if (toSend)
        delete [] toSend;
    
    return syncItem;
}

static bool setMailData(SyncItem &item, const wchar_t *parent, MailClientData &m, bool isUpdate)
{
    int errorCode; 

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = m.getEmailData()->parse(msg);
    
    //
    // -2 is a real error
    // -1 is an error only if it is a new item but not if it is an update.
    // 
    if (errorCode == 0 || (errorCode == -1 && isUpdate)) {
        // nothing to do.
        // it continue usually
    } else {
        LOG.error("Error parsing message!");
        return false;
    }
    m.setParent(parent);

    delete [] msg;
    return true;
}

static bool setFolderData(SyncItem &item, /*const wchar_t *parent*/ FolderData &folder)
{
    int errorCode; 

    char *msg = new char[item.getDataSize() + 1];
    memset(msg, 0, item.getDataSize() + 1);
    memcpy(msg, item.getData(), item.getDataSize());

    errorCode = folder.parse(msg);

    if (errorCode != 0) {
        LOG.error("Error parsing message!");
        return false;
    }
    
    delete [] msg;
    return true;
}


// set all extra data
//
static inline void setExtraData(ArrayList& extras,
                                const wchar_t *id,
                                const char *msgid, 
                                const char * extraHeaders)
{
    MailExtraData med;

    char *entryID = toMultibyte(id);
    med.setId(entryID);
    med.setMessageId(msgid);
    med.setHeaders(extraHeaders);
    extras.add(med);
    delete [] entryID;
}


MailSyncSource::MailSyncSource(const wchar_t* name) : SyncSource(name)
{    
    path = wstrdup(TEXT("\\"));
    s_folderToSync = NULL;   
    syncItem = NULL;
    isMailInOutbox = false;
    isMailInInbox = false;
    mailClient2Server = NULL;
    mailServer2Client = NULL;
    failedSendMailInOutbox = false;
}

MailSyncSource::MailSyncSource(MailSyncSource& s) : SyncSource(s)
{
    setLastSync(s.getLastSync());
    setNextSync(s.getNextSync());
    setRemoteURI(s.getRemoteURI());
    setType(s.getType());
    setLastAnchor(s.getLastAnchor());
    setNextAnchor(s.getNextAnchor());
    setErrorHandler(s.getErrorHandler());
    setFilter(s.getFilter());

    setPath(s.path);
    setFolderToSync(s.s_folderToSync);

    // These are not copied from the other instance
    isMailInOutbox = false;
    isMailInInbox = false;
    failedSendMailInOutbox = false;
    mailServer2Client=NULL;
    mailClient2Server=NULL;
    extras.clear();
}

MailSyncSource::~MailSyncSource() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;
    }
    if (s_folderToSync) {
        delete [] s_folderToSync; s_folderToSync = NULL;
    }
    if (path) {
        delete [] path; path = NULL;
    }
    if (mailServer2Client)
        delete mailServer2Client;
    if (mailClient2Server)
        delete mailClient2Server;
}

void MailSyncSource::setPath(const wchar_t* p) {
    if (path)
        delete [] path;

    if (p == NULL)
        path = wstrdup(TEXT(""));
    else
        path = wstrdup(p);
}

const wchar_t* MailSyncSource::getPath() {
    return path;
}

void MailSyncSource::setFolderToSync(const wchar_t* t){
    if(s_folderToSync)
        delete [] s_folderToSync;
    s_folderToSync = wstrdup(t);
}

const wchar_t* MailSyncSource::getFolderToSync() {
    return s_folderToSync;
}

bool MailSyncSource::isInFolderToSync(const wchar_t* toCheck) {
    return wcsstr(s_folderToSync, toCheck) != NULL ? true : false;
}

void MailSyncSource::setIsMailInOutbox(bool value){
    isMailInOutbox = value;
    
}

bool MailSyncSource::getIsMailInOutbox() {
    return isMailInOutbox;
}

void MailSyncSource::setIsMailInInbox(bool value){
    isMailInInbox = value;    
}

bool MailSyncSource::getIsMailInInbox() {
    return isMailInInbox;
}

void MailSyncSource::setFailedSendMailInOutbox(bool value){
    failedSendMailInOutbox = value;    
}

bool MailSyncSource::getFailedSendMailInOutbox() {
    return failedSendMailInOutbox;
}

/*
 * Return the first SyncItem of all. It is used in case of slow 
 * sync and retrieve the entire data source content.
 */
SyncItem* MailSyncSource::getFirstItem()
{
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    isMailInInbox = false;

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        LOG.error("getFirstItem: error in MailClient2Server::setSync");
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstItem();
    
    if (m) {              
        LOG.info(T("Subject first all mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());   
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NONE);
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        LOG.info(T("No mail get first"));
    }
    
    return syncItem;    
}

/*
* Return the next SyncItem of all. It is used in case of slow or refresh sync 
* and retrieve the entire data source content.
*/

SyncItem* MailSyncSource::getNextItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextItem();   
    if (m) {      
        LOG.info(T("Subject other all mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);                        
        syncItem->setState(SYNC_STATE_NONE);
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
                
        delete m; m = NULL;

    } else {
        LOG.info(T("No mail get next"));
    }

    return syncItem;  

}

SyncItem* MailSyncSource::getFirstItemKey() {
        
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    } 
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        LOG.error("getFirstItemKey: error in MailClient2Server::setSync");
        return NULL;
    }

    MailClientData* m = mailClient2Server->getFirstItem(TRUE);
    
    if (m) {
        mailServer2Client->setFolderToSync(s_folderToSync);
        mailServer2Client->deleteMessage(m);        
        delete m; m = NULL;

        while((m = mailClient2Server->getNextItem(TRUE)) != NULL) {
            mailServer2Client->deleteMessage(m);
            LOG.debug("Delete all items: one record deleted");
            delete m; m = NULL;
        }
    }

    return NULL;
}

SyncItem* MailSyncSource::getNextItemKey() {

    LOG.debug("MailSyncSource::getNextItemKey");
    return NULL;
}

/*
* This method is always called by the sync manager. So the setModifiedItemsPOOM is always called
*/

SyncItem* MailSyncSource::getFirstNewItem() {

    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_TWO_WAY, getPath(), extras) != S_OK ){
        LOG.error(T("getFirstNewItem: error in MailClient2Server::setSync"));
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstNewItem();   
    
    if (m) {      
        LOG.info(T("Subject first new mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NEW);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        LOG.info(T("No new mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextNewItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextNewItem();   
    
    if (m) {      
        LOG.info(T("Subject other new mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_NEW);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        delete m; m = NULL;

    } else {
        LOG.info(T("No other new mail"));
    }
    return syncItem;  
}

SyncItem* MailSyncSource::getFirstUpdatedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
        
    MailClientData* m = mailClient2Server->getFirstUpdatedItem();       

    if (m) {
        if(!m->getEmailData()->getEmailItem().empty()){
            LOG.info("Subject first updated mail: -%s-",
                m->getEmailData()->getEmailItem().getSubject());    
        }
        else {
            LOG.info("update flags only");            
        }

        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_UPDATED);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        delete m; m = NULL;

    } else {
        LOG.info(T("No updated mail"));
    }
    return syncItem;  
}

SyncItem* MailSyncSource::getNextUpdatedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextUpdatedItem();   
    
    if (m) {      
        LOG.info(T("Subject other updated mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());    
        syncItem = buildSyncItem(m, extras);        
        syncItem->setState(SYNC_STATE_UPDATED);
        //syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        delete m; m = NULL;

    } else {
        LOG.info(T("No other updated mail"));
    }
    /*
    if (m) {      
        
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            char *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }

        wchar_t *id = toWideChar(emailItem.getEntryID());
        syncItem = new SyncItem(id);
        delete [] id;
        

        char* toSend = m->getEmailData()->format();
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));
        delete [] toSend;
        syncItem->setState(SYNC_STATE_UPDATED);
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No other updated mail"));
    }
    */
    return syncItem;  
}

SyncItem* MailSyncSource::getFirstDeletedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getFirstDeletedItem();   
    
    if (m) {        
        syncItem = new SyncItem(m->getEntryId());
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        
        LOG.info(T("ID of deleted mail: -%S-"), m->getEntryId());
        delete m; m = NULL;        
    
    } else {
        LOG.info(T("No deleted mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextDeletedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextDeletedItem();   
    
    if (m) {
        syncItem = new SyncItem(m->getEntryId());     
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        LOG.info(T("ID of other deleted mail: -%S-"), m->getEntryId());

        delete m; m = NULL;

    } else {
        LOG.info(T("No other deleted mail"));
    }

    return syncItem;  
}

void MailSyncSource::setItemStatus(const wchar_t* key, int status) {
    
    if (wcsstr(key, TEXT("O/")) != NULL) {
        if (status == 500 || status == 418 || status == 419) {
            failedSendMailInOutbox = true;            
        }
    }

    LOG.info(T("Item status - key: %S, status: %i"), key, status);
}

int MailSyncSource::addItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    if (!isMemoryAvailable()) {
        LOG.error(LOW_DEVICE_MEMORY);
        return ret;
    }

    MailClientData m(TEXT("undef"));
    wchar_t* id = NULL;
    wchar_t tt[3];
    char* posFolder = NULL;
    mailServer2Client->setFolderToSync(s_folderToSync);
    
    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());
    }
    
    char* itemData = (char*)item.getData();
    if(!itemData){
        LOG.error("addItem: empty data from server");
        ret = 500;
        goto finally;
    }

    // Check Folder tag
    posFolder = strstr(itemData, "Folder>");
    if ((posFolder != NULL) && ((posFolder - itemData) < 10)) {
        FolderData folder;
        if (!setFolderData(item, folder))
        {
            ret = 500;
            goto finally;
        }
        else
        {
            ret = 200;
            if (wcsicmp(folder.getRole(), TEXT("Inbox")) == 0)
                item.setKey(TEXT("I"));
            else if (wcsicmp(folder.getRole(), TEXT("Outbox")) == 0)
                item.setKey(TEXT("O"));
            else if (wcsicmp(folder.getRole(), TEXT("Drafts")) == 0)
                item.setKey(TEXT("D"));
            else if (wcsicmp(folder.getRole(), TEXT("Sent")) == 0)
                item.setKey(TEXT("S"));
            else
                ret = 500;            
            goto finally;
        }
    }
    else{
        if (!isInFolderToSync(tt))
            goto finally;
        if (!setMailData(item, tt, m, false))
            goto finally;
    }

    id = mailServer2Client->addMessage(&m);
    if (!id) {
        LOG.error( T("Error in MailServer2Client::addMessage") );
        goto finally;
    }
    else if (wcscmp(id, ITEM_ARLEADY_EXISTS) == 0) {
        ret = STC_ALREADY_EXISTS; 
        LOG.debug( T("Mail arleady exists"));
        goto finally;
    }

    // Ok: set the new key and the status code
    ret = STC_ITEM_ADDED;        
    item.setKey(id);
    
    // commented only for debug. It must be used!!!!
    // MailMessage& msg = m.getEmailData()->getEmailItem();
    
    
    // logInfo(TEXT("Mail added: -%S-"), msg.getSubject());
    
    // Set new mail in Inbox flag
    if (isMailInInbox == false && wcscmp(tt, TEXT("I")) == 0) 
        isMailInInbox = true;

finally:
    if (id) { delete [] id; id = NULL; }
    return ret;
}

int MailSyncSource::updateItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    if (!isMemoryAvailable()) {
        LOG.error(LOW_DEVICE_MEMORY);
        return ret;
    }
    wchar_t tt[3];
    MailClientData m(item.getKey());; 
    int errorCode = 0;
    HRESULT hr = S_OK;
    char* tmp = NULL;
    wchar_t* toUse = NULL;

    mailServer2Client->setFolderToSync(s_folderToSync);
    
    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());    
    }
    /*
    char *id = toMultibyte(item.getKey());
    m.emailData->getEmailItem().setEntryID(id);
    delete [] id;
    */    

    if (!isInFolderToSync(tt))
        goto finally;     
    
    if (!setMailData(item, tt, m, true))
        goto finally;
           
    hr = mailServer2Client->updateMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;     
        LOG.info(T("Mail updated: -%s-"), m.getEmailData()->getEmailItem().getSubject());
    }    
    
finally:
    
    return ret;
}

int MailSyncSource::deleteItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    wchar_t tt[3];

    mailServer2Client->setFolderToSync(s_folderToSync);
    MailClientData m(item.getKey());

    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());    
    }

    m.setParent(tt);
    
    HRESULT hr = mailServer2Client->deleteMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;          
        LOG.info("Mail deleted: -%S-", item.getKey());
    }
        
    return ret;
}

int MailSyncSource::beginSync() {
    // Init mailServer2Client
    if (mailServer2Client)
        delete mailServer2Client;

    mailServer2Client = new MailServer2Client();
    
    if(!mailServer2Client){
        LOG.error("memory error getting mailServer2Client");
        return 500;
    }

    // Init mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
        
    if(!mailClient2Server){
        LOG.error("memory error getting mailServer2Client");
        return 500;
    }
    
    if (getSyncMode() == SYNC_REFRESH_FROM_SERVER) {
        int val = TimedMessageBox(NULL, ASK_REFRESH_MESSAGE, TEXT("Alert"),
                             MB_YESNO | MB_ICONQUESTION, ASK_REFRESH_TIMEOUT * 1000); // default ASK_TIMEOUT = 10

        if (val == IDYES) {
            LOG.debug("User said yes to delete local mail and refresh.");            
        }
        else {
            LOG.debug("User said no to delete local mail and refresh");     
            return -1;
        }
    }
    return 0;
}

int MailSyncSource::endSync() {
    
    // Get a new mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
    mailClient2Server->setSync(ONLY_ALL_MESSAGE_LIST, getPath(), extras);    
    mailClient2Server->writeCurrentMails(getPath(), extras);

    return 0;
}

MailSyncSource& MailSyncSource::operator=(MailSyncSource& s) {
    if (this != &s) {
        setSyncMode(s.getSyncMode());
        setLastSync(s.getLastSync());
        setNextSync(s.getNextSync());
        setRemoteURI(s.getRemoteURI());
        setType(s.getType());
        setLastAnchor(s.getLastAnchor());
        setNextAnchor(s.getNextAnchor());
        setErrorHandler(s.getErrorHandler());

        setFilter(s.getFilter());
        setPath(s.getPath());
        setFolderToSync(s.getFolderToSync());

        // This attributes are not copied from the other source
        setIsMailInOutbox(false);
        setIsMailInInbox(false);
        mailClient2Server = NULL;
        mailServer2Client = NULL;
        extras.clear();
    }
    return *this;
}

ArrayElement* MailSyncSource::clone() {
    MailSyncSource* s = new MailSyncSource(getName());
    *s = *this;
    return s;
}



