/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include "worker.h"
#include "s4jproxy.h"
#include "logutils.h"
#include "notify/util.h"
// Worker thread
extern "C" DWORD WINAPI NotifWorker(LPVOID lpv);

static bool syncinprogress = false;

// Handle of the worker thread
static HANDLE handle = 0;
// Connection socket
static SOCKET sock = 0;
static S4JProxy* s4jp = 0;
// Message buffer
static char *message = 0;
static int msglen = 0;
// Thread start time
time_t startTime;

static void addBuffer(const char *buf, int len);
static void closeWorker();

const char *getMsg() { return message; }
int getLen() { return msglen; }

/** Get the current time as Unix time_t */
static time_t getTime()
{
    SYSTEMTIME st;
    FILETIME ft;

    GetSystemTime(&st);
    SystemTimeToFileTime(&st, &ft);
    // Convert FILETIME to time_t
    __int64 llTmp;
    memcpy (&llTmp, &ft, sizeof (__int64));
    llTmp = (llTmp - 116444736000000000) / 10000000;
    return (time_t) llTmp;
}

DWORD startWorker(SOCKET s)
{
    // Check worker
    if (handle) {
        DebugPrint(TEXT("Another worker handle: %x\n"), handle);
        DWORD exitcode = 0;
        if(GetExitCodeThread(handle, &exitcode)) {
            if (exitcode == STILL_ACTIVE) {
                DebugPrint(TEXT("Worker thread still active\n"));
                if (syncinprogress) {
                    DebugPrint(TEXT("The sync is in progress\n"));
                    DebugPrint(TEXT("S4ND: connection busy\n"));
                    return ERROR_BUSY;
                }
                else if( (startTime - getTime()) <  ASK_TIMEOUT ) {
                    DebugPrint(TEXT("Request too close, wait\n"));
                    DebugPrint(TEXT("S4ND: connection busy\n"));
                    return ERROR_BUSY;
                }
            }
        }
        DebugPrint(TEXT("The sync is NOT in progress, closing thread\n"));
        if(sock){
            closesocket(sock);
		    sock= 0;
        }
        Sleep(100);
        if(handle){
            DebugPrint(TEXT("Closing handle\n"));
            CloseHandle(handle);
            handle=0;
        }
    }
    // Clean-up dirty messages
    if(message){
        delete [] message;
        message=0;
    }
    msglen=0;

    // Start worker
    sock = s;
	s4jp = S4JProxy::getInstance();

    startTime = getTime();   // record starting time
    handle = CreateThread(NULL, 0, NotifWorker, (LPVOID*) sock, 0, NULL);
    if (handle == 0) {
        DebugPrint(L"S4ND:ERROR:CreateThread fails."
                L" GetLastError() returns <0x%08x>\n",
                GetLastError());
        // Release resources
        closeWorker();
    }

    DebugPrint(L"S4ND: new worker thread handle=<0x%08x>\n", handle);
    return ERROR_SUCCESS;
}

DWORD stopWorker()
{
	if(sock){
        closesocket(sock);
		sock= 0;
    }
    Sleep(100);
    if ( !handle )
	    return ERROR_SUCCESS;
    else
        return ERROR_BUSY;
}

static void closeWorker()
{
    if(sock){
        closesocket(sock);
		sock= 0;
    }
    if(handle){
        CloseHandle(handle);
        handle=0;
    }
	S4JProxy::dispose();
    if(message){
        delete [] message;
        message=0;
    }
    msglen=0;

    DebugPrint(L"Worker end.\n");
	DebugEnd();
}

// Concatenate a char buf with specified len to the message buffer
static void addBuffer(const char *buf, int len)
{
    if(!buf || !len)
        return;

    if(!message){
        message = new char[len+1];
        memcpy(message, buf, len);
        msglen = len;
    }
    else{
        char *tmp = new char[msglen+len+1];
        memcpy(tmp, message, msglen);
        memcpy(tmp+msglen, buf, len);
        msglen += len;
        delete [] message;
        message = tmp;
    }
}

// Process the incoming message and return a response
// Return NULL if the message is not complete
static const char *processPacket(const char *pkt, int pktlen)
{
    S4JProxy::RetCode code = S4JProxy::Error;
    static time_t last = getTime();

    time_t now = getTime();

    if ( (now - last) > 5 ){
        // reset message buffer
        delete [] message;
        message = 0;
    }

    // Message arrived
    DebugPrint(L"S4ND: %d bytes received\n", pktlen);

    addBuffer(pkt, pktlen);

    DebugPrint(L"S4ND: total message len %d\n", msglen);
    HexDump(message, msglen);

    code=s4jp->parsePkg0(message, msglen);
    DebugPrint(L"Return from parseMessage: %d\n", code);//XXX

    if (code != S4JProxy::Continue)
        return s4jp->getResponse();
    else
        return NULL;
}

// Connection thread
extern "C" DWORD WINAPI NotifWorker(LPVOID lpv)
{
    SOCKET sock = (SOCKET)lpv;
    char buf[1501];
    int buflen=1500, pktlen;
    const char *response=NULL;

    while (!response) {
        pktlen = recv(sock, buf, buflen, 0);
        DebugPrint(L"Pktlen: %d\n", pktlen);    //XXX
        switch(pktlen) {
            case SOCKET_ERROR: {
                DWORD err = WSAGetLastError();
                if(err == WSAEMSGSIZE ){
                    // TODO: this case should be handled 
                    DebugPrint(L"Message larger than %d bytes!", buflen);
                    response = "500 - Message too long\n";
                }
                else {
                    DebugPrint(L"Recv error: %d", err);
                    response = "500 - Network error\n"; // Try to respond anyway?
                }
                break;
            }
            case 0:
                // End of connection
                DebugPrint(L"Connection closed by peer.\n");
                response = "";  // Conn closed, no response
                break;
            default:
				response = processPacket(buf, pktlen);
        }
    }

    // If message is not empty, send it
    if(*response){
        DebugPrint(L"S4ND: Sending message <%S> to client\r\n", response);
        send(sock, response, strlen(response),0);
	    // Start Sync
        if ( !s4jp->hasErrors() ) {
            closesocket(sock); // XXX
            syncinprogress=true;
	        s4jp->sync();
            DebugPrint(L"Sync End.\n");
            syncinprogress=false;
        }
    }
    // Delete thread
	closeWorker();
    return 0;
}


