/*
 * Copyright (C) 2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// startsync.cpp - main module for the sync application
//

// Windows Header Files:
#include <windows.h>
#include <initguid.h>
#include <Notify.h>
#include <Pmpolicy.h>

#include "base/startcmd.h"
#include "base/util/utils.h"

#include "pim/SettingFunctions.h"
#include "pim/maincpp.h"
#include "notify/addresschange.h"
#include "pim/difftime.h"
#include "usernotify.h"
#include "base/Log.h"
#include "pushmessage.h"
#include "notify/util.h"
#define DIMCONF 128

//--------------------------------------------------------------------- Prototypes

static inline void Alert(const wchar_t *msg, const wchar_t *title=TEXT("Warning"));

static int getPollingTime();
static bool isServerNotified();
static void setServerNotified(bool val);

//static int polling(const wchar_t **sources);
static int schedule(const wchar_t **sources);
static int sync(const wchar_t **sources);

static bool g_debug = true;


//--------------------------------------------------------------- Static Functions

static inline void Alert(const wchar_t *msg, const wchar_t *title)
{
#ifdef DEBUG
  	MessageBox (NULL, msg, title, MB_SETFOREGROUND |MB_OK);
#endif
}

static int getPollingTime()
{
    wchar_t buf[32];
    
    memset(buf, 0, 32*sizeof(wchar_t));
    getClientConfigurationInternal (TEXT(""), PROPERTY_SPDM_POLLING_NOTIFICATION, 
                                 buf, FALSE);
    return wcstol(buf, NULL, 10);
}

static bool isServerNotified()
{
    wchar_t buf[10];

    memset(buf, 0, 10*sizeof(wchar_t));
    getClientConfigurationInternal (TEXT(""), PROPERTY_SPDM_SERVER_NOTIFIED, 
                                     buf, FALSE);
    return ( wcscmp(buf, TEXT("1")) == 0);
        
}

static void setServerNotified(bool val)
{
    setClientConfiguration(
            TEXT(""),
            PROPERTY_SPDM_SERVER_NOTIFIED,
            ( (val) ? TEXT("1") : TEXT("0") ),
            NULL);
}

#if 0
/**
 * Start polling on the specified sources
 */
static int polling(const wchar_t **sources=NULL)
{    
    long t = getPollingTime();
    int ret = 0;

    if (t > 0 && !isServerNotified()) {
        LOG.debug("sync");
        ret = sync(sources);
        LOG.debug("sleep");
        Sleep(t*1000*60);
        LOG.debug("wakeup");
        t = getPollingTime();
        //
        // reschedule and exit.
        //
        if (t > 0) {
            wchar_t cmdLine[512];
            memset(cmdLine, 0, 512*sizeof(wchar_t));
            wcscpy(cmdLine, TEXT("poll "));
            
            for (int i = 0; sources[i]; i++) {
                wcscat(cmdLine, sources[i]);
                wcscat(cmdLine, TEXT(","));
            }
            cmdLine[wcslen(cmdLine) -1] = NULL;

            startcmd(SYNCAPP, cmdLine);
        }        
    }    
    return ret;

}
#endif

static void getAppName(wchar_t* appName) {
    const WCHAR* path = getProgramPath();
    wcscpy(appName, path);
    wcscat(appName, TEXT("\\"));
    wcscat(appName, SYNCAPP);
}

static int schedule(const wchar_t **sources=NULL)
{
    long t = getPollingTime();
    int ret = 0;
    LOG.debug("Schedule start");
    if (t > 0 && !isServerNotified()) {
        LOG.debug("sync");
        ret = sync(sources);
        // Re-schedule for next sync
        WCHAR appName[MAX_PATH];
        getAppName(appName);
        t = getPollingTime();
        if (t > 0){
            SYSTEMTIME lpTime;
            GetLocalTime(&lpTime);
            SYSTEMTIME pollTime = DT_AddDiff(nano100SecInMin, t, &lpTime);
            if (CeRunAppAtTime(appName,&pollTime))
                ret = 0;
            else
                ret = -1;
        }
    }
    LOG.debug("Schedule end");
    return ret;
}



/**
 * Syncronize the specified sources
 */
static int sync(const wchar_t **sources=NULL)
{
    HANDLE hMutex;//Handle to mutex...

    // Create mutex here
    hMutex = CreateMutex(NULL, TRUE, TEXT("FunSyncInProgress") );

    switch(GetLastError()) {
        case ERROR_SUCCESS:
            LOG.debug("Mutex created.");
            break;  
        case ERROR_ALREADY_EXISTS:    
            if (g_debug)
                Alert(TEXT("Another sync in progress"), TEXT("Warning"));
            LOG.error("Another sync in progress");
            return ERR_ANOTHER_SYNC;
        default:
            LOG.error("Failed to create mutex");
            return -1;
    }
    
    const wchar_t *path = getProgramPath();
    wchar_t errmsg[1024];   
    HANDLE handleBubble = CreateThread(NULL, 0, createBubble, NULL, 0, NULL);
    if (handleBubble == 0) {
        LOG.error("CreateThread failed: <0x%08x>\n", GetLastError());
    }
    WaitForSingleObject(handleBubble, 0);

    int ret = synchronize(errmsg, path, (const wchar_t **)sources);
    
    removeNotif();
        
    CloseHandle( hMutex );
    LOG.debug("Mutex released.");

    return ret;
}

static int notify(const wchar_t **sources=NULL)
{
    AN_ResponseCode responseCode = notifyAddressChange(TEXT(APPLICATION_URI));

    if(responseCode == AN_AddressAccepted ){
        setServerNotified(true);
    }
    else {
        setServerNotified(false);
        // Start the polling, if the polling time is set.
        schedule(sources);
    }
    return 0;
}

/**
 * Parse the command line
 *
 * @param cmdline - the command line (syntax: command [source,source,...])
 *
 * @param command (out) - the command name
 * @param sources (out) - a NULL terminated array of source names
 */
static int parseCmdLine(const LPTSTR cmdline,
                        wchar_t **command,
                        wchar_t ***sources)
{
    if( wcsstr(cmdline, TEXT("debug ")) ) {
        *command = wstrdup(cmdline+wcslen(TEXT("debug ")));
        g_debug = true;
    }
    else {
        *command = wstrdup(cmdline);
        g_debug = false;
    }

    *sources = NULL;

    wchar_t *args = wcschr(*command, ' ');
    if(!args) {
        // in this case comand == cmdline and sources is null
        return 0;
    }
    // terminate command and point to the start of arguments
    args[0] = 0; args++;
    if (!args[0]) {
        // empty args after a space, like before
        return 0;
    }
    int argc;
    wchar_t *arg = args;
    // Count arguments
    for(argc = 1; (arg=wcschr(arg+1, ',')) != 0; argc++);

    // Get arg list
    wchar_t **srclist = new wchar_t *[argc+1];
    arg=args;
	int i;
    for(i=0; i<argc; i++) {
        if(!arg || !arg[0])
            return -1;
        srclist[i]=arg;
 
        arg=wcschr(arg, ',');
        if(arg) {
            arg[0]=0; arg++;
        }
    }
    srclist[i]=0;
    *sources=srclist;

    return 0;
}


//-------------------------------------------------------------------------- MAIN
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow) {    
    
    
    wchar_t *command, **sources;
    int ret = 0;
    HRESULT         hr              = NOERROR;
    HANDLE          hevtMsgAvailable= NULL;
    HPUSHROUTER     hPushRouter     = NULL;
    PUSHMSG         PushMsg         = {0};

    PowerPolicyNotify(PPN_UNATTENDEDMODE, TRUE);
    //SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_BELOW_NORMAL);

    LOG.setLevel(LOG_LEVEL_DEBUG);

    if (parseCmdLine(lpCmdLine, &command, &sources)) {
        Alert(TEXT("Syntax error in command line"), TEXT("StartSync"));
        return 1;
    }

    if(g_debug){
        Alert(lpCmdLine);
    }

    if(wcscmp(command, TEXT("poll")) == 0) {
        ret = schedule((const wchar_t **)sources);
    }
    else if (wcscmp(command, TEXT("once")) == 0) {
        ret = sync((const wchar_t **)sources);
    }
    else if (wcscmp(command, TEXT("addresschange")) == 0) {
        ret = notify((const wchar_t **)sources);
    }
    else if (wcscmp(lpCmdLine, APP_RUN_AT_TIME) == 0) {
        const wchar_t *syncSources[] = { TEXT("mail"), NULL };
        ret = schedule(syncSources);        
    }
    //
    // register the startsync to the pushrouter
    //
    else if(wcscmp(lpCmdLine, TEXT("register")) == 0) { 
                
        WCHAR c_szPath[MAX_PATH];
        getAppName(c_szPath); // /Program files/Funambol/startsync.exe

        hr = PushRouter_RegisterClient(c_szContentType, c_szAppId, c_szPath, c_szParams);
        if (hr == S_OK) {            
            LOG.debug("Wap push client registered successfully");
        } else if (hr == PUSHRTR_E_ALREADYREGISTERED) {
            LOG.debug("Wap push client arleady registered");
        } else {
            LOG.error("Error to register wap push client");
        }   
    }
    //
    // unregister the startsync from the pushrouter
    //
    else if(wcscmp(lpCmdLine, TEXT("deregister")) == 0) {         
        hr = PushRouter_UnRegisterClient(c_szContentType, c_szAppId);  
        
        if (hr == S_OK) {            
            LOG.debug("Wap push client unregistered successfully");
        } else if (hr == PUSHRTR_E_NOTFOUND) {
            LOG.debug("Wap push client not found. Arleady unregistered");
        } else {
            LOG.error("Error to unregister wap push client");
        }
    }        
    else if(wcscmp(lpCmdLine, c_szParams) == 0) 
    {   
        memset(&PushMsg, 0, sizeof(PUSHMSG));
        PushMsg.cbSize = sizeof(PUSHMSG);

        hr = PushRouter_Open(c_szContentType, c_szAppId, &hevtMsgAvailable, &hPushRouter);
        if (hr != S_OK) {
            LOG.error("Wap push: PushRouter_Open - it is not possible to register the client.");
        }

        while(WaitForSingleObject(hevtMsgAvailable, 10000) != WAIT_TIMEOUT) 
        {
            hr = PushRouter_GetMessage(hPushRouter, &PushMsg);
            if (hr != S_OK) {
                LOG.error("Error in PushRouter_GetMessage wap push client");
                break;
            }  
            wchar_t** ss = NULL;
            int ret = handleWapPushMessage(&PushMsg, &ss);
            if (ret != 0) {
                LOG.error("Wap push: error in handleWapPushMessage message");
                break;
            }
            if (ss) {
                if (askConfirm(ASK_WAP_PUSH)) {
                    ret = sync((const wchar_t **)ss); 
                }
            }

            PushRouter_FreeMessage(&PushMsg);
            memset(&PushMsg, 0, sizeof(PUSHMSG));
            PushMsg.cbSize = sizeof(PUSHMSG);
        }
    }
    else {
        Alert(TEXT("Unknown command."), TEXT("StartSync"));        
    }
    
    if (command)
        delete [] command;
    if (sources)
        delete [] sources;

    if(g_debug) {
        wchar_t dbg[200];
        wsprintf(dbg, TEXT("Exit code: %d"), ret);
        Alert(dbg, TEXT("StartSync"));
    }
        
        
    if (hPushRouter) {
        PushRouter_Close(hPushRouter);
        PushRouter_FreeMessage(&PushMsg);
    }
    
    PowerPolicyNotify(PPN_UNATTENDEDMODE, FALSE);

	return ret;
}

